// ***********************************************************************************************************************************************
// File             : ISO14443A_2.c
// Version          : 1.1
// Written by       : Olexsandr Zhuk, Ruffieux Denis
// Project          : MLX90121 13.56MHz Transceiver
// Date             : 29.01.2007
// Company          : Melexis Technologies SA (Switzerland)
// Microcontroller  : ATmega128
// Description      : Routines for the communication protocol ISO14443A level 2.
// ***********************************************************************************************************************************************
#include "Defs.h"
#include "HwDefs.h"
#include "IsoX.h"
#include <avr\io.h>
#include "ErrorCodes.h"
#include "Crt.h"
#include "Mlx90121.h"

#define CIso14ARecvDelay 100

BYTE FIso14ALastSymbol;

void Iso14ASendBit ( BYTE AData )
{
 if (AData&0x01)        //All bits are represented by a sequence X, Y or Z (refer to ISO14443-2)
  {
   MlxSendNextSym1P(0);     //Send Sequence X which is represented by Sym0 of MLX90121 (refer to the datasheet MLX90121)
   FIso14ALastSymbol=1;	//Update Last bit Sent variable to '1'
  }
 else
  {
   if (FIso14ALastSymbol==0) MlxSendNextSym2P(2); // Z
   else MlxSendNextSym1P(1); // Y
   FIso14ALastSymbol=0; //Update Last bit Sent variable to '0'
  }
}

BYTE Iso14ASend ( BYTE *ABuf, BYTE AByteCount, BYTE ABitBuf, BYTE ABitCount, BYTE AAutoBcc, BYTE AAutoCrc )
{
 BYTE BByteIndex,
      BBitIndex;
 BYTE BData;   // Internal data byte
 BYTE BParity; // Internal Parity byte
 BYTE BResult;
 BYTE BBcc;
 WORD BCrc;
 BYTE BIntSave;

 BBcc=0;
 if ((AAutoBcc)&&(AByteCount>2))
  {
   for (BByteIndex=2; BByteIndex<AByteCount; BByteIndex++) BBcc^=ABuf[BByteIndex];
   ABuf[AByteCount]=BBcc; AByteCount++;
  }

 BCrc=0;
 if (AAutoCrc)
  {
   BCrc=Iso14AGetCrc(ABuf,AByteCount);
   ABuf[AByteCount]=BCrc; AByteCount++;
   ABuf[AByteCount]=BCrc>>8; AByteCount++;
  }

 FIso14ALastSymbol=0; // Reset Global variable for Last Symbol Sent.

 BIntSave=IntDisable();
 MlxSendStart(0x0B,127,2);

 for (BByteIndex=0; BByteIndex<AByteCount; BByteIndex++)
  {
   BData=*ABuf;  // Get next byte
   BParity=0x01; // Initialise Parity bit

   for (BBitIndex=0; BBitIndex<8; BBitIndex++) //Send byte bit-by-bit
    {
     BParity^=BData;       //Calculate Parity bit
     Iso14ASendBit(BData); //Send current bit of the byte
     BData>>=1;            //Shift data left once. Send data MSB first according to ISO14443A
    }

   Iso14ASendBit(BParity); //Send Parity bit

   ABuf++;
  }

 BData=ABitBuf;
 for (BBitIndex=0; BBitIndex<ABitCount; BBitIndex++)
  {
   Iso14ASendBit(BData); //Send current bit
   BData>>=1;            //Shift data left once. Send data MSB first according to ISO14443A
  }

 Iso14ASendBit(0); //Send EOF = '0' + Sequence Y
 MlxSendNextSym1P(1);

 BResult=MlxSendStop(0x0A,CIso14ARecvDelay);

 IntRestore(BIntSave);

 return BResult;
}

BYTE Iso14ARecv ( BYTE *ABuf, BYTE ABufSize, BYTE ABitsToSkip, BYTE *ABytesReceived, WORD *AFirstCollBit, BYTE *AParityTest )
{
 BYTE BResult;
 BYTE BIntSave;
 BYTE BDataA,
      BDataB,
      BDataC;
 BYTE BFirstCollByte,
      BFirstCollBit;
 BYTE *BBuf;
 BYTE BBitIndex,
      BByteCount;
 BYTE BParityTest;
 BYTE BParity;


 BResult=ENoTag;
 TrbRead; ModeComm;

 *ABytesReceived=0; *AFirstCollBit=0xFFFF; *AParityTest=FALSE;
 BBuf=ABuf; BParityTest=TRUE;
 BFirstCollByte=0xFF; BFirstCollBit=0x07; BByteCount=0;
 BIntSave=IntDisable();
 while (TimerTest==0);

 //PORTC=0x01;
 BDataA=0x00; BDataB=0xFF; BDataC=0x00; // Just for satisfy compiler

 do
  {
   if (MlxSyncStartASK(0x0A,127)==FALSE) break;

   //Prepare Timers for Timeout, Reset counters
   TimerSet(0x0A,127);
   BByteCount=0;

   //PORTC=0x03;
   TimerReset;
   while((DSYN!=0)&&(TimerTest==0));
   while((DSYN==0)&&(TimerTest==0));
   //PORTC=0x01;
   BDataA=GetData;

   //PORTC=0x03;
   TimerReset;
   while((DSYN!=0)&&(TimerTest==0));
   while((DSYN==0)&&(TimerTest==0));
   //PORTC=0x01;
   BDataB=GetData;

   if ((BDataA!=1)||(BDataB!=0)) { BResult=EInvalidSof; break; }

   BResult=EFrameWaitTOut;
   BBitIndex=ABitsToSkip; BByteCount=0; BDataC=0;
   do
    {
     // Receive byte bit by bit
     BParity=0x01;
     while (BBitIndex<8)
      {
       //PORTC=0x03;
       TimerReset;
       while((DSYN!=0)&&(TimerTest==0));
       while((DSYN==0)&&(TimerTest==0));
       //PORTC=0x01;
       BDataA=GetData;
       BParity^=BDataA;
       BDataC=(BDataC>>1)|(BDataA<<7);

       if (BBitIndex==7)
        {
         if (BByteCount>=ABufSize) { BResult=EBufOverflow; break; }
         *BBuf=BDataC;
        }

       //PORTC=0x03;
       TimerReset;
       while((DSYN!=0)&&(TimerTest==0));
       while((DSYN==0)&&(TimerTest==0));
       //PORTC=0x01;
       if (BDataA==GetData) // Collision
        {
         if (BDataA==0) // End of communication?
          {
           if (BBitIndex==0) BResult=EFrameRecvOK;
           else BResult=EFramingError;
           break;
          }
         else // Real collision
          {
           if (BFirstCollByte==0xFF) { BFirstCollByte=BByteCount; BFirstCollBit=BBitIndex; }
          }
        }
       BBitIndex++;
      } // while
     if (BBitIndex!=8) break;
     BBuf++; BByteCount++;
     BBitIndex=0;
     if (TimerTest) { BResult=ENoDSYN; break; }

     // receive parity
     //PORTC=0x03;
     TimerReset;
     while((DSYN!=0)&&(TimerTest==0));
     while((DSYN==0)&&(TimerTest==0));
     //PORTC=0x01;
     BDataA=GetData;
     if ((ABitsToSkip!=0)&&(BByteCount==1))
      {
       if (BParity!=BDataA) BParityTest=FALSE;
      }

     //PORTC=0x03;
     TimerReset;
     while((DSYN!=0)&&(TimerTest==0));
     while((DSYN==0)&&(TimerTest==0));
     //PORTC=0x01;
     if (BDataA==GetData) // Collision. Do not analize type of collision here, because it complicates the things
      {
       if (BFirstCollByte==0xFF) { BFirstCollByte=BByteCount; BFirstCollBit=0; }
      }

    } while(1);

  } while(0);

 MlxSyncStop();
 IntRestore(BIntSave);

 //PORTC=0x00;

 *AParityTest=BParityTest;
 if (BResult==EFrameRecvOK)
  {
   *ABytesReceived=BByteCount;
   if (BFirstCollByte!=0xFF) *AFirstCollBit=(BFirstCollByte<<3)+BFirstCollBit;
  }

 return BResult;
}


