#include "Defs.h"
#include "Crt.h"
#include "IsoX.h"
#include "Iso14A.h"
#include "ErrorCodes.h"

/* We can have maximum 96 bits of UID (32*3=96). UID itself we store in FIso14AInvUid buffer.
   For example, we determined 16 bits of UID and 17th bit is collided. So, the first
   16 bits of FIso14AInvUid we keep unchanged. The 17th bit can be 0 or 1 (or both). We store
   '17' in a stack and start to resolve UID assuming that bit 17 = 0. Of course, now
   we can have a collision in bit 18, 19 and so on, we will repeat this procedure,
   increasing stack. Later we will come back to bit 17 and will try to resolve UID,
   assuming that bit 17 = 1. */

BYTE FIso14AInvCollStack[96];
BYTE FIso14AInvCsIndex;
BYTE FIso14AInvUid[12];
BYTE FIso14AInvFlags;

//#define Iso14AInvDebug

#if defined Iso14AInvDebug

#include <avr\pgmspace.h>
#include "Comm.h"

void Iso14AParTagReport ( BYTE AResult, BYTE *ABuf, BYTE ABytesReceived, WORD AFirstCollBit, BYTE AParityTest );

/* Functions to send debug information to the PC */
void Iso14AInvViewSend ( BYTE *AComment, BYTE *ADataBuf, BYTE ADataSize, BYTE ABitBuf, BYTE ABitCount )
{
 BYTE i;

 CommSendStrC(AComment);
 CommSendStrC(FlashConst(" "));
 for (i=0; i<ADataSize; i++)
  {
   if (i>=2) break;
   CommSendHex(ADataBuf[i]);
  }
 CommSendStrC(FlashConst(" "));
 for (; i<ADataSize; i++)
  {
   if (i>=6) break;
   CommSendHex(ADataBuf[i]);
  }
 CommSendStrC(FlashConst(" "));
 for (; i<ADataSize; i++)
  {
   CommSendHex(ADataBuf[i]);
  }

 if (ABitCount)
  {
   CommSendHex(ABitBuf);
   CommSendStrC(FlashConst(" "));
   CommSendHex(ABitCount);
  }

 CommSendStrC13(FlashConst(""));
}

#endif

/* Function shows which bit is set in byte. For example,
   if AColl = 0000 0100 it returns 2. Useful to determine the
   number of collided bit */
BYTE GetCollIndex ( BYTE AColl )
{
 BYTE BResult;

 BResult=0;
 while (AColl)
  {
   BResult++;
   if (AColl&0x01) break;
   AColl>>=1;
  }

 return (BResult);
}

/* Function to set or reset a bit in byte stream.
   If before calling
       *ADataBuf = 00000000 00000000 00000000 ...
       ABitCount = 17
       AData     = 1
   then, after calling
       *ADataBuf = 00000000 00000000 00000010 ... */

void Iso14AInvSetBit ( BYTE *ADataBuf, BYTE ABitCount, BYTE AData )
{
 BYTE BByteIndex, BBitIndex;
 BYTE ABitIndex;

 ABitIndex=ABitCount-1;
 BByteIndex=ABitIndex/8;
 BBitIndex=ABitIndex-BByteIndex*8;
 ADataBuf[BByteIndex]&=(0xFF>>(7-BBitIndex));
 if (AData) ADataBuf[BByteIndex]|=(0x01<<BBitIndex);
 else ADataBuf[BByteIndex]&=((0x01<<BBitIndex)^0xFF);
}

/* Iso14AInvSendRecv sends a request to the TAG according to the ISO-norm.
     ALevel     - anticollision level, from 0 to 2. It is used because UID can be
                  long: 4, 7 or 10 bytes and full UID must be read in few etapes.
     AUid       - pointer to a 4 bytes UID which we want to send as anticollision
     ABitCount  - the number of bit we send
     AClearBits - the number of clear (non-collided) bits in the response

   After execution the decoded response will be copied into AUid buffer.

   Request command depends of ALevel.
   ALevel  ->  0  1  2
   Command ->  93 95 97

   if we send a request for Level 1, ABitCount = 17, *AUid = 880485EB, the bit
   stream will look like this
   Bit 11001001 P 10000100 P 00010001 P 00100000 P 1
   HEX    93         21         88         04      1 (the low part of 85)

   If ABitCount = 32 (i.e. we send a complete SELECT command) the command will
   be different: it must include BCC and CRC. The response will also be different
*/

BYTE Iso14AInvSendRecv ( BYTE ALevel, BYTE *AUid, BYTE ABitCount, BYTE *AClearBits )
{
 BYTE BDataS[12];
 BYTE BResult;
 BYTE BByteCount, BBitCount, i, j;
 BYTE BMask;
 BYTE BBytesReceived;
 WORD BFirstCollBit;
 BYTE BParityTest;
 BYTE BBitToSkip;

 BResult=ENoTag;
 *AClearBits=0;

 /* determine byte and bit count */
 BByteCount=ABitCount/8;
 BBitCount=ABitCount-BByteCount*8;
 BMask=0xFF>>(8-BBitCount);

 /* fill-in the buffer to send */
 BDataS[0]=0x93+(ALevel<<1);
 if (ABitCount==32) BDataS[1]=0x70;
 else BDataS[1]=0x20+(BByteCount<<4)+BBitCount;
 for (i=0; i<BByteCount; i++) BDataS[2+i]=AUid[i];
 if (i<4) { BDataS[2+i]=AUid[i] & BMask; i++; }
 for ( ; i<4; i++) BDataS[2+i]=0;

 /* if ABitCount = 32, append BCC and CRC */
 /* Note that sending (Iso14AH is also slightly different */
 if (ABitCount==32)
  {
  #if defined Iso14AInvDebug
   Iso14AInvViewSend((BYTE *)FlashConst("SELECT 70"),&BDataS[0],6,0,0);
  #endif
   BBitToSkip=0;
   Iso14ASend(&BDataS[0],6,0,0,TRUE,TRUE);
  }
 else
  {
  #if defined Iso14AInvDebug
   Iso14AInvViewSend((BYTE *)FlashConst("SELECT"),&BDataS[0],BByteCount+2,BDataS[2+BByteCount],BBitCount);
  #endif
   BBitToSkip=BBitCount;
   Iso14ASend(&BDataS[0],BByteCount+2,BDataS[2+BByteCount],BBitCount,FALSE,FALSE);
  }

 do /* while(0) - error loop */
  {
   BResult=Iso14ARecv(&BDataS[0],12,BBitToSkip,&BBytesReceived,&BFirstCollBit,&BParityTest);
  #if defined Iso14AInvDebug
   Iso14AParTagReport(BResult,&BDataS[0],BBytesReceived,BFirstCollBit,BParityTest);
  #endif
   if (BResult!=EFrameRecvOK) break; /* didn't receive anything? Exit */

   /* If we sent 32 bits, the check mechanism is different */
   if (ABitCount==32)
    {
     if (BBytesReceived!=3) { BResult=EFramingError; break; }
     *AClearBits=32;
     break;
    }

   i=BByteCount; j=0;           
   if (i<4)
    {
     AUid[i]=(AUid[i]&BMask)|(BDataS[j]&(BMask^0xFF));
     i++; j++;
    }
   while (i<4)
    {
     AUid[i]=BDataS[j];
     i++; j++;
    }

   if (BFirstCollBit==0xFFFF) *AClearBits=32;
   else *AClearBits=BFirstCollBit;
  } while (0);

 return (BResult);
}

void Iso14AInvSelectOpti ( void )
{
 BYTE BBitCount;

 Iso14AInvSendRecv(0,&FIso14AInvUid[0],32,&BBitCount);
}

/* Iso14AInvSendReqA sends REQA request. It doesn't decode the answer */
BYTE Iso14AInvSendReqA ( void )
{
 BYTE BDataS[12];
 BYTE BResult;
 BYTE BBytesReceived;
 WORD BFirstCollBit;
 BYTE BParityTest;

#if defined Iso14AInvDebug
 CommSendStrC13(FlashConst("REQA (Rollback)  <-"));
#endif
 Iso14ASend(0,0,0x26,7,FALSE,FALSE);
 BResult=Iso14ARecv(&BDataS[0],12,0,&BBytesReceived,&BFirstCollBit,&BParityTest);
#if defined Iso14AInvDebug
 Iso14AParTagReport(BResult,&BDataS[0],BBytesReceived,BFirstCollBit,BParityTest);
#endif

 return BResult;
}

/* Iso14AInvSendWupA sends WUPA request. It doesn't decode the answer */
BYTE Iso14AInvSendWupA ( void )
{
 BYTE BDataS[12];
 BYTE BResult;
 BYTE BBytesReceived;
 WORD BFirstCollBit;
 BYTE BParityTest;

#if defined Iso14AInvDebug
 CommSendStrC13(FlashConst("WUPA (Start)"));
#endif
 Iso14ASend(0,0,0x52,7,FALSE,FALSE);
 BResult=Iso14ARecv(&BDataS[0],12,0,&BBytesReceived,&BFirstCollBit,&BParityTest);
#if defined Iso14AInvDebug
 Iso14AParTagReport(BResult,&BDataS[0],BBytesReceived,BFirstCollBit,BParityTest);
#endif

 return BResult;
}

/* Iso14AInvSendHltA sends HltA request. It doesn't decode the answer */
BYTE Iso14AInvSendHltA ( void )
{
 BYTE BDataS[12];
 BYTE BResult;
 BYTE BBytesReceived;
 WORD BFirstCollBit;
 BYTE BParityTest;

#if defined Iso14AInvDebug
 CommSendStrC13(FlashConst("HALT"));
#endif
 BDataS[0]=0x50;
 BDataS[1]=0x00;
 Iso14ASend(&BDataS[0],2,0,0,FALSE,TRUE);
 BResult=Iso14ARecv(&BDataS[0],12,0,&BBytesReceived,&BFirstCollBit,&BParityTest);
#if defined Iso14AInvDebug
 Iso14AParTagReport(BResult,&BDataS[0],BBytesReceived,BFirstCollBit,BParityTest);
#endif

 return BResult;
}

/* Iso14AInvRollBack performs a rollback till the value in a stack (see above) till
   the value in the stack matches ALevel */
void Iso14AInvRollBack ( BYTE ALevel )
{
 while (FIso14AInvCsIndex>0)
  {
   if ((FIso14AInvCollStack[FIso14AInvCsIndex-1]/32)<ALevel) break;
   FIso14AInvCsIndex--;
  }
}

/* The main function Iso14AInvGetNext obtains the next UID based on values in the stack */
BYTE Iso14AInvGetNext ( BYTE *ABuf )
{
 BYTE BBitCount,
      BBitCountA;
 BYTE BResult;
 BYTE BResultA;
 BYTE BLevel;
 BYTE BKeepSelected;
 BYTE i;

 BResult=ENoTag;

 BKeepSelected=FALSE;
 do
  {
   if ((FIso14AInvCsIndex==0) && (FIso14AInvFlags==0)) break; /* Top of the stack */
   /* FIso14AInvFlags shows the direction. 0 - decrease stack (get postponed value) */
   /* 1 - increase. Postpone value to resolve it later */
   if (FIso14AInvFlags==0)
    {
     Iso14AInvSendReqA();

     FIso14AInvFlags=0x01;
     FIso14AInvCsIndex--;
     BBitCount=FIso14AInvCollStack[FIso14AInvCsIndex];
     BBitCount++;
     Iso14AInvSetBit(&FIso14AInvUid[0],BBitCount,1);
     FIso14AInvCollStack[FIso14AInvCsIndex]=BBitCount;
    }

   do
    {
     FIso14AInvFlags=0; /* Rollback by default */

     BBitCountA=FIso14AInvCollStack[FIso14AInvCsIndex];
     BLevel=BBitCountA/32;
     /* if necessary, send select commands for all previous levels to reactivate TAG */
     if ((BLevel>0)&&(BKeepSelected==FALSE))
      {
       BResultA=Iso14AInvSendRecv(0,&FIso14AInvUid[0],32,&BBitCount);
       if (BResultA!=EFrameRecvOK) { Iso14AInvRollBack(0); break; }
      }
     if ((BLevel>1)&&(BKeepSelected==FALSE))
      {
       BResultA=Iso14AInvSendRecv(1,&FIso14AInvUid[4],32,&BBitCount);
       if (BResultA!=EFrameRecvOK) { Iso14AInvRollBack(1); break; }
      }

     /* try to send BBitCountA bits */
     BResultA=Iso14AInvSendRecv(BLevel,&FIso14AInvUid[BLevel*4],BBitCountA%32,&BBitCount);
     if (BResultA!=EFrameRecvOK) break;

     if (BBitCount==32) /* there is no collision -> TAG is found */
      {
       BKeepSelected=TRUE;

       /* perform a check of the found TAG */
       BResultA=Iso14AInvSendRecv(BLevel,&FIso14AInvUid[BLevel*4],32,&BBitCount);
       if ((BResultA!=EFrameRecvOK)||(BBitCount!=32)) break;
       /* check if there is a next part of UID (next level) */
       if ((FIso14AInvUid[BLevel*4]!=0x88)||(BLevel==2))
        {
         for (i=0; i<12; i++) ABuf[i]=FIso14AInvUid[i];
         BResult=EFrameRecvOK;
         break;
        }

       BBitCount=(BLevel+1)*32;
       FIso14AInvFlags=1;
       FIso14AInvCollStack[FIso14AInvCsIndex]=BBitCount;
       break;
      }

     /* Branch this bit. On the next step try to set collided bit to 0, and postpone
        the checking with the bit = 1 */
     BKeepSelected=FALSE;
     BBitCount=BBitCount+BLevel*32;
     FIso14AInvFlags=1;

     if (FIso14AInvCollStack[FIso14AInvCsIndex]>BBitCount) BBitCount=FIso14AInvCollStack[FIso14AInvCsIndex];
     else FIso14AInvCollStack[FIso14AInvCsIndex]=BBitCount;
     FIso14AInvCsIndex++;
     BBitCount++;
     Iso14AInvSetBit(&FIso14AInvUid[0],BBitCount,0);
     FIso14AInvCollStack[FIso14AInvCsIndex]=BBitCount;
    } while (0);

   if (BResult==EFrameRecvOK) break;

  } while (1);

 return (BResult);
}

/* Function Iso14AInvGetFirstUid initialises the start values and calls Iso14AInvGetNextUid */
void Iso14AInvStart ( void )
{
 BYTE i;

 for (i=0; i<12; i++) FIso14AInvUid[i]=0;
 FIso14AInvCollStack[0]=0;
 FIso14AInvCsIndex=0;
 FIso14AInvFlags=0x01;
 Iso14AInvSendWupA();
}


